import threading

from ._iomediator import IoMediator


class Waiter:
    """
    Waits for a specified event to occur
    """

    def __init__(self, rawKey, modifiers, button, timeOut):
        IoMediator.listeners.append(self)
        self.rawKey = rawKey
        self.modifiers = modifiers
        self.button = button
        self.event = threading.Event()
        self.timeOut = timeOut

        if modifiers is not None:
            self.modifiers.sort()

    def wait(self):
        return self.event.wait(self.timeOut)

    def handle_keypress(self, rawKey, modifiers, key, *args):
        if rawKey == self.rawKey and modifiers == self.modifiers:
            IoMediator.listeners.remove(self)
            self.event.set()

    def handle_mouseclick(self, rootX, rootY, relX, relY, button, windowInfo):
        if button == self.button:
            self.event.set()



import re

from .key import Key

X_RECORD_INTERFACE = "XRecord"
KEY_SPLIT_RE = re.compile("(<[^<>]+>\+?)")

MODIFIERS = [Key.CONTROL, Key.ALT, Key.ALT_GR, Key.SHIFT, Key.SUPER, Key.HYPER, Key.META, Key.CAPSLOCK, Key.NUMLOCK]
HELD_MODIFIERS = [Key.CONTROL, Key.ALT, Key.SUPER, Key.SHIFT, Key.HYPER, Key.META]



import re
import unittest

from lib.phrase import *

class PhraseTest(unittest.TestCase):
    
    def setUp(self):
        # Set up global settings
        """globals = {
                    WORD_CHARS_REGEX_OPTION : re.compile('[\w]', re.UNICODE),
                    IMMEDIATE_OPTION : False,
                    IGNORE_CASE_OPTION : False,
                    MATCH_CASE_OPTION : False,
                    BACKSPACE_OPTION : True,
                    OMIT_TRIGGER_OPTION : False,
                    TRIGGER_INSIDE_OPTION : False
                   }"""
        
        self.defaultPhrase = Phrase("xp@", "expansion@autokey.com")
        self.defaultPhrase.set_abbreviation("xp@")
        self.defaultPhrase.set_modes([PhraseMode.ABBREVIATION])
        self.defaultFolder = PhraseFolder("Folder")
        self.defaultFolder.add_phrase(self.defaultPhrase)        
        
    def testImmediateOption(self):
        
        # Test default setting (false)
        self.assertEqual(self.defaultPhrase.check_input("xp@", ""), False)
        self.assertEqual(self.defaultPhrase.calculate_input("xp@ "), 4)
        
        # Test true setting
        phrase = Phrase("xp@", "expansion@autokey.com")
        phrase.immediate = True
        phrase.set_abbreviation("xp@")
        phrase.set_modes([PhraseMode.ABBREVIATION])
        self.defaultFolder.add_phrase(phrase)
        
        self.assertEqual(phrase.check_input("xp@", ""), True)
        result = phrase.build_phrase("xp@")
        self.assertEqual(result.string, "expansion@autokey.com")
        self.assertEqual(result.backspaces, 3)
        self.assertEqual(phrase.calculate_input("xp@"), 3)
        
        
    def testIgnoreCaseOption(self):
        
        # Test default setting (false)
        self.assertEqual(self.defaultPhrase.check_input("XP@ ", ""), False)
        
        # Test true setting
        phrase = Phrase("xp@", "expansion@autokey.com")
        phrase.ignoreCase = True
        phrase.set_abbreviation("xp@")
        phrase.set_modes([PhraseMode.ABBREVIATION])
        self.defaultFolder.add_phrase(phrase)
        self.assertEqual(phrase.check_input("XP@ ", ""), True)
        
        result = phrase.build_phrase("XP@ ")
        self.assertEqual(result.string, "expansion@autokey.com ")
        
    def testMatchCaseOption(self):
        phrase = Phrase("xp@", "expansion@autokey.com")
        phrase.ignoreCase = True
        phrase.matchCase = True
        phrase.set_abbreviation("xp@")
        phrase.set_modes([PhraseMode.ABBREVIATION])        
        self.defaultFolder.add_phrase(phrase)
        
        result = phrase.build_phrase("asdf XP@ ")
        self.assertEqual(result.string, "EXPANSION@AUTOKEY.COM ")        
        
        result = phrase.build_phrase("ASDF Xp@ ")
        self.assertEqual(result.string, "Expansion@autokey.com ")
        
        result = phrase.build_phrase("Asdf xp@ ")
        self.assertEqual(result.string, "expansion@autokey.com ")        
        
    def testBackspaceOption(self):        
        # Test default setting (true)
        result = self.defaultPhrase.build_phrase("xp@ ")
        self.assertEqual(result.backspaces, 4)
        
        # Test false setting
        phrase = Phrase("xp@", "expansion@autokey.com")
        phrase.backspace = False
        phrase.set_abbreviation("xp@")
        phrase.set_modes([PhraseMode.ABBREVIATION])        
        self.defaultFolder.add_phrase(phrase)
        
        result = phrase.build_phrase("xp@ ")
        self.assertEqual(result.backspaces, 1)
        
    def testOmitTriggerOption(self):        
        # Test default setting (false)
        result = self.defaultPhrase.build_phrase("xp@\n")
        self.assertEqual(result.string, "expansion@autokey.com\n")
        
        # Test true setting
        phrase = Phrase("xp@", "expansion@autokey.com")
        phrase.omitTrigger = True
        phrase.set_abbreviation("xp@")
        phrase.set_modes([PhraseMode.ABBREVIATION])        
        self.defaultFolder.add_phrase(phrase)
        
        result = phrase.build_phrase("xp@.")
        self.assertEqual(result.string, "expansion@autokey.com")
        self.assertEqual(result.backspaces, 4)
        
    def testTriggerInsideOption(self):        
        # Test default setting (false)
        self.assertEqual(self.defaultPhrase.check_input("asdfxp@\n", ""), False)

        # when separated by a non-word char, should still trigger
        self.assertEqual(self.defaultPhrase.check_input("asdf.xp@ ", ""), True)
        
        # Test true setting
        phrase = Phrase("xp@", "expansion@autokey.com")
        phrase.triggerInside = True
        phrase.set_abbreviation("xp@")
        phrase.set_modes([PhraseMode.ABBREVIATION])        
        self.defaultFolder.add_phrase(phrase)
        self.assertEqual(phrase.check_input("asdfxp@.", ""), True)
        
        result = phrase.build_phrase("asdfxp@.")
        self.assertEqual(result.string, "expansion@autokey.com.")
    
    def testLefts(self):
        #Test with omit trigger false
        phrase = Phrase("Positioning Phrase", "[udc]$(cursor )[/udc]")
        phrase.set_abbreviation("udc")
        phrase.set_modes([PhraseMode.ABBREVIATION])
        self.defaultFolder.add_phrase(phrase)
        
        result = phrase.build_phrase("udc ")
        phrase.parsePositionTokens(result)
        self.assertEqual(result.lefts, 7)
        
        #Test with omit trigger true
        phrase.omitTrigger = True
        result = phrase.build_phrase("udc ")
        phrase.parsePositionTokens(result)
        self.assertEqual(result.lefts, 6)
        
        #Test with immediate true
        phrase.omitTrigger = False
        phrase.immediate = True
        result = phrase.build_phrase("udc")
        phrase.parsePositionTokens(result)
        self.assertEqual(result.lefts, 6)
        
    def testLeftsWithSpecialKey(self):
        phrase = Phrase("Positioning Phrase", "[udc]$(cursor )[/udc]<f1>")
        phrase.set_abbreviation("udc")
        phrase.set_modes([PhraseMode.ABBREVIATION])
        self.defaultFolder.add_phrase(phrase)
        
        result = phrase.build_phrase("udc ")
        phrase.parsePositionTokens(result)
        self.assertEqual(result.lefts, 7)        
        
    def testMultipleAbbrs(self):
        phrase = Phrase("Some abbr", "Some abbr")
        phrase.set_abbreviation("sdf")
        phrase.set_modes([PhraseMode.ABBREVIATION])        
        self.defaultFolder.add_phrase(phrase)
        input = "fgh xp@asdf sdf "
        
        # Abbreviation should not trigger
        self.assertEqual(self.defaultPhrase.check_input(input, ""), False)
        
        self.assertEqual(phrase.check_input(input, ""), True)
        
    def testSort(self):
        phrase1 = Phrase("abbr1", "Some abbr")
        phrase2 = Phrase("abbr2", "Some abbr")
        self.defaultFolder.add_phrase(phrase1)
        self.defaultFolder.add_phrase(phrase2)
        
        phrases = [phrase1, phrase2]
        
        phrase1.usageCount = 0
        phrase2.usageCount = 1

        phrases.sort(reverse=True)

        self.assertEqual(phrases[0].description, "abbr2")
        self.assertEqual(phrases[1].description, "abbr1")
        
    def testNoneMode(self):
        phrase = Phrase("Test Phrase", "Testing")
        folder = PhraseFolder("Folder")
        folder.set_abbreviation("asdf")
        folder.set_modes([PhraseMode.ABBREVIATION])
        folder.add_phrase(phrase)
        result = phrase.build_phrase("asdf ")
        
        self.assertEqual(result.backspaces, 5)
        self.assertEqual(result.string, "Testing")
        self.assertEqual(phrase.calculate_input("asdf "), 5)
        self.assertEqual(phrase.should_prompt("asdf "), False)
        
    def testWindowName(self):
        phrase = Phrase("Some abbr", "Some abbr")
        phrase.set_window_titles(".*Eclipse.*")
        phrase.set_abbreviation("sdf")
        phrase.set_modes([PhraseMode.ABBREVIATION])                
        self.defaultFolder.add_phrase(phrase)
        self.assertEqual(phrase.check_input("sdf ", "blah - Eclipse Platform"), True)


class PredictivePhraseTest(unittest.TestCase):
    
    def setUp(self):
        self.phrase = Phrase("blah", "This is a test phrase")
        self.phrase.set_modes([PhraseMode.PREDICTIVE])
        folder = PhraseFolder("Folder")
        folder.add_phrase(self.phrase)
        
    def testPredict(self):
        self.assertEqual(self.phrase.check_input("This ", ""), True)
        self.assertEqual(self.phrase.check_input("This", ""), False)
        self.assertEqual(self.phrase.check_input("This i", ""), False)
        self.assertEqual(self.phrase.check_input("I don't have a problem. This ", ""), True)
        
    def testBuildPhrase(self):
        result = self.phrase.build_phrase("This ")
        self.assertEqual(result.backspaces, 0)
        self.assertEqual(result.string, "is a test phrase")
        
    def testCalcInput(self):
        self.assertEqual(self.phrase.calculate_input("This "), 5)
        
    def testShouldPrompt(self):
        self.assertEqual(self.phrase.should_prompt("This "), True)
        
class HotkeyPhrasetest(unittest.TestCase):
    
    def setUp(self):
        self.phrase = Phrase("blah", "This is a test phrase")
        self.phrase.set_modes([PhraseMode.HOTKEY])
        self.phrase.set_hotkey(["A", "B"], "n")
        folder = PhraseFolder("Folder")
        folder.add_phrase(self.phrase)
        
    def testHotkey(self):
        result = self.phrase.check_hotkey(["A", "B"], "n", "")
        self.assertEqual(result, True)
        
        result = self.phrase.check_hotkey(["B"], "n", "")
        self.assertEqual(result, False)
        
        result = self.phrase.check_hotkey(["A", "B"], "a", "")
        self.assertEqual(result, False)
        
    def testBuildPhrase(self):
        result = self.phrase.build_phrase("")
        self.assertEqual(result.string, "This is a test phrase") 
        
    def testCalcInput(self):
        self.assertEqual(self.phrase.calculate_input(''), 3)
        
class PhraseFolderTest(unittest.TestCase):
    
    def setUp(self):
        self.folder = PhraseFolder("Folder")
        self.folder.set_abbreviation("sdf")
        self.folder.set_modes([PhraseMode.ABBREVIATION])
                
    def testCheckInput(self):
        self.assertEqual(self.folder.check_input("sdf ", ""), True)
    
class E2ETest(unittest.TestCase):
    
    def setUp(self):
        self.topFolder = PhraseFolder("Top Folder")
        self.topFolder.set_abbreviation("top1")
        self.topFolder.set_modes([PhraseMode.ABBREVIATION])
        
        self.bottomFolder = PhraseFolder("Bottom Folder")
        self.bottomFolder.set_abbreviation("bottom1")
        self.bottomFolder.set_modes([PhraseMode.ABBREVIATION])
        self.topFolder.add_folder(self.bottomFolder)
        
        self.phrase = Phrase("blah", "The Phrase")
        self.phrase.set_abbreviation("asdf")
        self.phrase.set_hotkey(["A"], "n")
        self.phrase.set_modes([PhraseMode.ABBREVIATION, PhraseMode.HOTKEY, PhraseMode.PREDICTIVE])
        self.bottomFolder.add_phrase(self.phrase)
            
    def testCheckInput(self):
        self.assertEqual(self.topFolder.check_input("top1 ", ""), True)
        self.assertEqual(self.bottomFolder.check_input("bottom1 ", ""), True)
        self.assertEqual(self.phrase.check_input("asdf ", ""), True)
        self.assertEqual(self.phrase.check_input("The P", ""), True)
        self.assertEqual(self.phrase.check_hotkey(["A"], "n", ""), True)
        
    def testBuildPhraseDirect(self):
        result = self.phrase.build_phrase("asdf ")
        self.assertEqual(result.string, "The Phrase ")
        self.assertEqual(result.backspaces, 5)
        
        result = self.phrase.build_phrase("The P")
        self.assertEqual(result.string, "hrase")
        self.assertEqual(result.backspaces, 0)
        
        result = self.phrase.build_phrase("")
        self.assertEqual(result.string, "The Phrase")
        self.assertEqual(result.backspaces, 0)
        
    def testBackspacesIndirect(self):
        result = self.phrase.build_phrase("top1 ")
        self.assertEqual(result.backspaces, 5)
        
        result = self.phrase.build_phrase("bottom1 ")
        self.assertEqual(result.backspaces, 8)




# -*- coding: utf-8 -*-

# Copyright (C) 2011 Chris Dekter
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import common
common.USING_QT = False

import sys
import traceback
import os.path
import signal
import logging
import logging.handlers
import subprocess
import optparse
import time
import threading

import gettext
import dbus
import dbus.service
import dbus.mainloop.glib

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject, GLib

gettext.install("autokey")


from autokey import service, monitor
from autokey.gtkui.notifier import get_notifier
from autokey.gtkui.popupmenu import PopupMenu
from autokey.gtkui.configwindow import ConfigWindow
from autokey import configmanager as cm

PROGRAM_NAME = _("AutoKey")  # TODO: where does this _ named function come from? It must be one of those from x import *
DESCRIPTION = _("Desktop automation utility")
COPYRIGHT = _("(c) 2008-2011 Chris Dekter")


class Application:
    """
    Main application class; starting and stopping of the application is controlled
    from here, together with some interactions from the tray icon.
    """

    def __init__(self):
        GLib.threads_init()
        Gdk.threads_init()

        p = optparse.OptionParser()
        p.add_option("-l", "--verbose", help="Enable verbose logging", action="store_true", default=False)
        p.add_option("-c", "--configure", help="Show the configuration window on startup", action="store_true", default=False)
        options, args = p.parse_args()

        try:
            # Create configuration directory
            if not os.path.exists(common.CONFIG_DIR):
                os.makedirs(common.CONFIG_DIR)
            # Create data directory (for log file)
            if not os.path.exists(common.DATA_DIR):
                os.makedirs(common.DATA_DIR)
            # Create run directory (for lock file)
            if not os.path.exists(common.RUN_DIR):
                os.makedirs(common.RUN_DIR)

            # Initialise logger
            rootLogger = logging.getLogger()

            if options.verbose:
                rootLogger.setLevel(logging.DEBUG)
                handler = logging.StreamHandler(sys.stdout)
            else:
                rootLogger.setLevel(logging.INFO)
                handler = logging.handlers.RotatingFileHandler(common.LOG_FILE,
                                        maxBytes=common.MAX_LOG_SIZE, backupCount=common.MAX_LOG_COUNT)

            handler.setFormatter(logging.Formatter(common.LOG_FORMAT))
            rootLogger.addHandler(handler)

            if self.__verifyNotRunning():
                self.__createLockFile()

            self.initialise(options.configure)

        except Exception as e:
            self.show_error_dialog(_("Fatal error starting AutoKey.\n") + str(e))
            logging.exception("Fatal error starting AutoKey: " + str(e))
            sys.exit(1)

    def __createLockFile(self):
        with open(common.LOCK_FILE, "w") as lock_file:
            lock_file.write(str(os.getpid()))

    def __verifyNotRunning(self):
        if os.path.exists(common.LOCK_FILE):
            pid = Application._read_pid_from_lock_file()

            # Check that the found PID is running and is autokey
            with subprocess.Popen(["ps", "-p", pid, "-o", "command"], stdout=subprocess.PIPE) as p:
                output = p.communicate()[0]

            if "autokey" in output.decode():
                logging.debug("AutoKey is already running as pid %s", pid)
                bus = dbus.SessionBus()

                try:
                    dbusService = bus.get_object("org.autokey.Service", "/AppService")
                    dbusService.show_configure(dbus_interface="org.autokey.Service")
                    sys.exit(0)
                except dbus.DBusException as e:
                    logging.exception("Error communicating with Dbus service")
                    self.show_error_dialog(_("AutoKey is already running as pid %s but is not responding") % pid, str(e))
                    sys.exit(1)

        return True

    @staticmethod
    def _read_pid_from_lock_file() -> str:
        with open(common.LOCK_FILE, 'r') as lock_file:
            return lock_file.read()

    def initialise(self, configure):
        logging.info("Initialising application")
        self.monitor = monitor.FileMonitor(self)
        self.configManager = cm.get_config_manager(self)
        self.service = service.Service(self)
        self.serviceDisabled = False

        # Initialise user code dir
        if self.configManager.userCodeDir is not None:
            sys.path.append(self.configManager.userCodeDir)

        try:
            self.service.start()
        except Exception as e:
            logging.exception("Error starting interface: " + str(e))
            self.serviceDisabled = True
            self.show_error_dialog(_("Error starting interface. Keyboard monitoring will be disabled.\n" +
                                    "Check your system/configuration."), str(e))

        self.notifier = get_notifier(self)
        self.configWindow = None
        self.monitor.start()

        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.dbusService = common.AppService(self)

        if configure:
            self.show_configure()

    def init_global_hotkeys(self, configManager):
        logging.info("Initialise global hotkeys")
        configManager.toggleServiceHotkey.set_closure(self.toggle_service)
        configManager.configHotkey.set_closure(self.show_configure_async)

    def config_altered(self, persistGlobal):
        self.configManager.config_altered(persistGlobal)
        self.notifier.rebuild_menu()

    def hotkey_created(self, item):
        logging.debug("Created hotkey: %r %s", item.modifiers, item.hotKey)
        self.service.mediator.interface.grab_hotkey(item)

    def hotkey_removed(self, item):
        logging.debug("Removed hotkey: %r %s", item.modifiers, item.hotKey)
        self.service.mediator.interface.ungrab_hotkey(item)

    def path_created_or_modified(self, path):
        time.sleep(0.5)
        changed = self.configManager.path_created_or_modified(path)
        if changed and self.configWindow is not None:
            self.configWindow.config_modified()

    def path_removed(self, path):
        time.sleep(0.5)
        changed = self.configManager.path_removed(path)
        if changed and self.configWindow is not None:
            self.configWindow.config_modified()

    def unpause_service(self):
        """
        Unpause the expansion service (start responding to keyboard and mouse events).
        """
        self.service.unpause()
        self.notifier.update_tool_tip()

    def pause_service(self):
        """
        Pause the expansion service (stop responding to keyboard and mouse events).
        """
        self.service.pause()
        self.notifier.update_tool_tip()

    def toggle_service(self):
        """
        Convenience method for toggling the expansion service on or off.
        """
        if self.service.is_running():
            self.pause_service()
        else:
            self.unpause_service()

    def shutdown(self):
        """
        Shut down the entire application.
        """
        if self.configWindow is not None:
            if self.configWindow.promptToSave():
                return

            self.configWindow.hide()

        self.notifier.hide_icon()

        t = threading.Thread(target=self.__completeShutdown)
        t.start()

    def __completeShutdown(self):
        logging.info("Shutting down")
        self.service.shutdown()
        self.monitor.stop()
        Gdk.threads_enter()
        Gtk.main_quit()
        Gdk.threads_leave()
        os.remove(common.LOCK_FILE)
        logging.debug("All shutdown tasks complete... quitting")

    def notify_error(self, message):
        """
        Show an error notification popup.

        @param message: Message to show in the popup
        """
        self.notifier.notify_error(message)

    def update_notifier_visibility(self):
        self.notifier.update_visible_status()

    def show_configure(self):
        """
        Show the configuration window, or deiconify (un-minimise) it if it's already open.
        """
        logging.info("Displaying configuration window")
        if self.configWindow is None:
            self.configWindow = ConfigWindow(self)
            self.configWindow.show()
        else:
            self.configWindow.deiconify()

    def show_configure_async(self):
        Gdk.threads_enter()
        self.show_configure()
        Gdk.threads_leave()

    def main(self):
        logging.info("Entering main()")
        Gdk.threads_enter()
        Gtk.main()
        Gdk.threads_leave()

    def show_error_dialog(self, message, details=None):
        """
        Convenience method for showing an error dialog.
        """
        dlg = Gtk.MessageDialog(type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK,
                                 message_format=message)
        if details is not None:
            dlg.format_secondary_text(details)
        dlg.run()
        dlg.destroy()

    def show_script_error(self, parent):
        """
        Show the last script error (if any)
        """
        if self.service.scriptRunner.error != '':
            dlg = Gtk.MessageDialog(type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK,
                                     message_format=self.service.scriptRunner.error)
            self.service.scriptRunner.error = ''
            # revert the tray icon
            self.notifier.set_icon(cm.ConfigManager.SETTINGS[cm.NOTIFICATION_ICON])
            self.notifier.errorItem.hide()
            self.notifier.update_visible_status()

        else:
            dlg = Gtk.MessageDialog(type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK,
                                     message_format=_("No error information available"))

        dlg.set_title(_("View script error"))
        dlg.set_transient_for(parent)
        dlg.run()
        dlg.destroy()

    def show_popup_menu(self, folders: list=None, items: list=None, onDesktop=True, title=None):
        if items is None:
            items = []
        if folders is None:
            folders = []
        self.menu = PopupMenu(self.service, folders, items, onDesktop, title)
        self.menu.show_on_desktop()

    def hide_menu(self):
        self.menu.remove_from_desktop()



import time
import os
import subprocess
import tempfile
import imghdr
import struct


class PatternNotFound(Exception):
    pass

# numeric representation of the mouse buttons. For use in visgrep.
LEFT = 1
MIDDLE = 2
RIGHT = 3


def visgrep(scr: str, pat: str, tolerance: int = 0) -> int:
    """
    visgrep(scr: str, pat: str, tolerance: int = 0) -> int
    Visual grep of scr for pattern pat.
    Requires xautomation (http://hoopajoo.net/projects/xautomation.html).
    visgrep("screen.png", "pat.png")
    Exceptions raised: ValueError, PatternNotFound, FileNotFoundError

    :param scr: path of PNG image to be grepped.
    :param pat: path of pattern image (PNG) to look for in scr.
    :param tolerance: An integer ≥ 0 to specify the level of tolerance for 'fuzzy' matches. If negative or not convertible to int, raises ValueError.
    :returns: coordinates of the topleft point of the match, if any. Raises PatternNotFound exception otherwise.
    """
    tol = int(tolerance)
    if tol < 0:
        raise ValueError("tolerance must be ≥ 0.")
    with open(scr), open(pat):
        pass
    with tempfile.NamedTemporaryFile() as f:
        subprocess.call(['png2pat', pat], stdout=f)
        # don't use check_call, some versions (1.05) have a missing return statement in png2pat.c so the exit status ≠ 0
        f.flush()
        os.fsync(f.fileno())
        vg = subprocess.Popen(['visgrep', '-t' + str(tol), scr, f.name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out = vg.communicate()
    coord_str = out[0].decode().split(' ')[0].split(',')
    try:
        coord = [int(coord_str[0]), int(coord_str[1])]
    except (ValueError, IndexError) as e:
        raise PatternNotFound(str([x.decode() for x in out]) + '\n\t' + repr(e))
    return coord


def get_png_dim(filepath: str) -> int:
    """
    get_png_dim(filepath:str) -> (int)
    Finds the dimension of a PNG.
    :param filepath: file path of the PNG.
    :returns: (width, height).
    """
    if not imghdr.what(filepath) == 'png':
        raise Exception("not PNG")
    head = open(filepath, 'rb').read(24)
    return struct.unpack('!II', head[16:24])


def mouse_move(x: int, y: int, display: str=''):
    subprocess.call(['xte', '-x', display, "mousemove {} {}".format(int(x), int(y))])


def mouse_rmove(x: int, y: int, display: str=''):
    subprocess.call(['xte', '-x', display, "mousermove {} {}".format(int(x), int(y))])


def mouse_click(button: int, display: str=''):
    subprocess.call(['xte', '-x', display, "mouseclick {}".format(int(button))])


def mouse_pos():
    tmp = subprocess.check_output("xmousepos").decode().split()
    return list(map(int, tmp))[:2]


def click_on_pat(pat: str, mousebutton: int=1, offset: (float, float)=None, tolerance: int=0, restore_pos: bool=False) -> None:
    """
    Requires imagemagick, xautomation, xwd.
    Click on a pattern at a specified offset (x,y) in percent of the pattern dimension. x is the horizontal distance from the top left corner, y is the vertical distance from the top left corner. By default, the offset is (50,50), which means that the center of the pattern will be clicked at.
    Exception PatternNotFound is raised when the pattern is not found on the screen.
    :param pat: path of pattern image (PNG) to click on.
    :param mousebutton: mouse button number used for the click
    :param offset: offset from the top left point of the match. (float,float)
    :param tolerance: An integer ≥ 0 to specify the level of tolerance for 'fuzzy' matches. If negative or not convertible to int, raises ValueError.
    :param restore_pos: return to the initial mouse position after the click.
    """
    x0, y0 = mouse_pos()
    move_to_pat(pat, offset, tolerance)
    mouse_click(mousebutton)
    if restore_pos:
        mouse_move(x0, y0)


def move_to_pat(pat: str, offset: (float, float)=None, tolerance: int=0) -> None:
    """See help for click_on_pat"""
    with tempfile.NamedTemporaryFile() as f:
        subprocess.call('''
        xwd -root -silent -display :0 | 
        convert xwd:- png:''' + f.name, shell=True)
        loc = visgrep(f.name, pat, tolerance)
    pat_size = get_png_dim(pat)
    if offset is None:
        x, y = [l + ps//2 for l, ps in zip(loc, pat_size)]
    else:
        x, y = [l + ps*(off/100) for off, l, ps in zip(offset, loc, pat_size)]
    mouse_move(x, y)


def acknowledge_gnome_notification():
    """
    Moves mouse pointer to the bottom center of the screen and clicks on it.
    """
    x0, y0 = mouse_pos()
    mouse_move(10000, 10000)  # TODO: What if the screen is larger? Loop until mouse position does not change anymore?
    x, y = mouse_pos()
    mouse_rmove(-x/2, 0)
    mouse_click(LEFT)
    time.sleep(.2)
    mouse_move(x0, y0)




import datetime
from abc import abstractmethod

from autokey.iomediator.constants import KEY_SPLIT_RE
from autokey.iomediator.key import Key
from autokey import common

if common.USING_QT:
    from PyQt5.QtWidgets import QAction

    def _(text: str, args: tuple=None):
        """localisation function, currently returns the identity. If args are given, those are used to format
        text using the old-style % formatting."""
        if args:
            text = text % args
        return text

    class MacroAction(QAction):

        def __init__(self, menu, macro, callback):
            super(MacroAction, self).__init__(macro.TITLE, menu)
            self.macro = macro
            self.callback = callback
            self.triggered.connect(self.on_triggered)

        def on_triggered(self):
            self.callback(self.macro)

else:
    from gi.repository import Gtk


class MacroManager:
    
    def __init__(self, engine):
        self.macros = []
        
        self.macros.append(ScriptMacro(engine))
        self.macros.append(DateMacro())
        self.macros.append(FileContentsMacro())
        self.macros.append(CursorMacro())
        
    def get_menu(self, callback, menu=None):
        if common.USING_QT:
            for macro in self.macros:
                menu.addAction(MacroAction(menu, macro, callback))
        
        else:
            menu = Gtk.Menu()

            for macro in self.macros:
                menuItem = Gtk.MenuItem(macro.TITLE)
                menuItem.connect("activate", callback, macro)
                menu.append(menuItem)

            menu.show_all()

        return menu
        
    def process_expansion(self, expansion):
        parts = KEY_SPLIT_RE.split(expansion.string)
        
        for macro in self.macros:        
            macro.process(parts)
        
        expansion.string = ''.join(parts)
        

class AbstractMacro:

    def get_token(self):
        ret = "<%s" % self.ID
        # TODO: v not used in initial implementation? This results in something like "<%s a= b= c=>"
        ret += "".join((" " + k + "=" for k, v in self.ARGS))
        ret += ">"
        return ret
            
    def _can_process(self, token):
        if KEY_SPLIT_RE.match(token):
            return token[1:-1].split(' ', 1)[0] == self.ID
        else:
            return False
        
    def _get_args(self, token):
        l = token[:-1].split(' ')
        ret = {}
                
        if len(l) > 1:
            for arg in l[1:]:
                key, val = arg.split('=', 1)
                ret[key] = val

        for k, v in self.ARGS:
            if k not in ret:
                raise Exception("Missing mandatory argument '{}' for macro '{}'".format(k, self.ID))
        
        return ret
        
    def process(self, parts):
        for i in range(len(parts)):
            if self._can_process(parts[i]):
                self.do_process(parts, i)

    @abstractmethod
    def do_process(self, parts, i):
        pass


class CursorMacro(AbstractMacro):

    ID = "cursor"
    TITLE = _("Position cursor")
    ARGS = []
    
    def do_process(self, parts, i):
        try:
            lefts = len(''.join(parts[i+1:]))
            parts.append(Key.LEFT * lefts)
            parts[i] = ''
        except IndexError:
            pass
                        
    
class ScriptMacro(AbstractMacro):

    ID = "script"
    TITLE = _("Run script")
    ARGS = [("name", _("Name")),
            ("args", _("Arguments (comma separated)"))]
    
    def __init__(self, engine):
        self.engine = engine
    
    def do_process(self, parts, i):
        args = self._get_args(parts[i])
        self.engine.run_script_from_macro(args)
        parts[i] = self.engine.get_return_value()


class DateMacro(AbstractMacro):

    ID = "date"
    TITLE = _("Insert date")
    ARGS = [("format", _("Format"))]
    
    def do_process(self, parts, i):
        format_ = self._get_args(parts[i])["format"]
        date = datetime.datetime.now().strftime(format_)
        parts[i] = date


class FileContentsMacro(AbstractMacro):

    ID = "file"
    TITLE = _("Insert file contents")
    ARGS = [("name", _("File name"))]
    
    def do_process(self, parts, i):
        name = self._get_args(parts[i])["name"]
        
        with open(name, "r") as inputFile:
            parts[i] = inputFile.read()



# Copyright (C) 2011 Chris Dekter
# Copyright (C) 2018 Thomas Hess <thomas.hess@udo.edu>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re

from PyQt5.QtWidgets import QDialog

from autokey.qtui import common as ui_common
from .detectdialog import DetectDialog

from autokey import iomediator
from autokey import model


# TODO: Once the port to Qt5 is done, enable the clearButtonEnable property for the line edit in the UI editor.
# TODO: Pure Qt4 does not support the line edit clear button, so this functionality is currently unavailable.
class WindowFilterSettingsDialog(*ui_common.inherits_from_ui_file_with_name("window_filter_settings_dialog")):

    def __init__(self, parent):
        super(WindowFilterSettingsDialog, self).__init__(parent)
        self.setupUi(self)
        self.target_item = None
        self.grabber = None  # type: iomediator.WindowGrabber

    def load(self, item: model.Item):
        self.target_item = item

        if not isinstance(item, model.Folder):
            self.apply_recursive_check_box.hide()
        else:
            self.apply_recursive_check_box.show()

        if not item.has_filter():
            self.reset()
        else:
            self.trigger_regex_line_edit.setText(item.get_filter_regex())
            self.apply_recursive_check_box.setChecked(item.isRecursive)

    def save(self, item):
        try:
            item.set_window_titles(self.get_filter_text())
        except re.error:
            # TODO: Warn user. Currently just doesn't save regex on error.
            pass
        item.set_filter_recursive(self.get_is_recursive())

    def get_is_recursive(self):
        return self.apply_recursive_check_box.isChecked()

    def reset(self):
        self.trigger_regex_line_edit.clear()
        self.apply_recursive_check_box.setChecked(False)

    def reset_focus(self):
        self.trigger_regex_line_edit.setFocus()

    def get_filter_text(self):
        return str(self.trigger_regex_line_edit.text())

    def receive_window_info(self, info):
        self.parentWidget().window().app.exec_in_main(self._receiveWindowInfo, info)

    def _receiveWindowInfo(self, info):
        dlg = DetectDialog(self)
        dlg.populate(info)
        dlg.exec_()

        if dlg.result() == QDialog.Accepted:
            self.trigger_regex_line_edit.setText(dlg.get_choice())

        self.detect_window_properties_button.setEnabled(True)

    # --- Signal handlers ---

    def on_detect_window_properties_button_pressed(self):
        self.detect_window_properties_button.setEnabled(False)
        self.grabber = iomediator.WindowGrabber(self)
        self.grabber.start()

    # --- event handlers ---

    def slotButtonClicked(self, button):
        if button == QDialog.Cancel:
            self.load(self.targetItem)

        QDialog.slotButtonClicked(self, button)

